#!/usr/bin/env python3
import sys
import argparse
import logging
from .config import read_config
from .plugins import load_plugins
from .logger import setup_logging
from .cache import Cache
from .errors import ConfigError
from .errors import NoResourcesFound

from . import __version__

# Setup logging
logger = logging.getLogger(__name__)


def main() -> None:
    """
    Main method.
    """
    parser = argument_parse()
    args = parser.parse_args()
    verbosity_level = verbosity(args.verbose)
    setup_logging(args.logger, verbosity_level)
    logger.info("Reading the rundeck-resources configuration")
    config = read_config(args.config)
    logger.info("Initializing cache system")
    cache = Cache(config, args.no_cache)
    logger.info("Loading importer plugins")
    try:
        import_plugins = load_importers(config)
    except ConfigError as e:
        logger.critical("Importers error: %s", e)
        sys.exit(1)
    logger.info("Loading exporter plugins")
    try:
        export_plugins = load_exporters(config)
    except ConfigError as e:
        logger.critical("Exporters error: %s", e)
        sys.exit(1)

    logger.info("Initializing input interfaces")
    logger.debug("Input plugins: %s", import_plugins)
    import_interfaces = initialize_import_interfaces(
        config, import_plugins, cache)
    logger.info("Initializing output interfaces")
    logger.debug("Output plugins: %s", export_plugins)
    export_interfaces = initialize_export_interfaces(
        config, export_plugins)

    logger.info("Importing nodes resources from input plugins")
    logger.debug("Input interfaces: %s", import_interfaces)
    try:
        resources = import_resources(import_interfaces)
    except NoResourcesFound as e:
        logger.error("Import error: %s", e)
        logger.warning("No resources file written")
        sys.exit(0)
    logger.info("Exporting nodes resources using output plugins")
    logger.debug("Output interfaces: %s", export_interfaces)
    export_resources(export_interfaces, resources)
    logger.info("Rundeck Resources successfully completed")


def load_importers(config):
    """
    Method to load the `Importers` plugins configured in the configuration
    file.

    :param config: The configuration file content.
    :returns: The `Importers` plugins, loaded.
    """
    return load_plugins(config, 'Importers')


def load_exporters(config):
    """
    Method to load the `Exporters` plugins configured in the configuration
    file.

    :param config: The configuration file content.
    :returns: The `Exporters` plugins, loaded.
    """
    return load_plugins(config, 'Exporters')


def initialize_import_interfaces(config: dict, plugins: dict,
                                 cache: Cache) -> list:
    """
    Method to initialize the interfaces with the configuration.

    :param config: The configuration file content.
    :param plugins: The list of loaded plugins.
    :param cache: The cache system instance.
    :returns: The list of loaded plugins initialized.
    """
    interfaces = []
    for interface in plugins.values():
        interfaces.append(
            interface['plugin'](interface['title'],
                                config, cache))
    return interfaces


def initialize_export_interfaces(config: dict, plugins: dict) -> list:
    """
    Method to initialize the interfaces with the configuration.

    :param config: The configuration file content.
    :param plugins: The list of loaded plugins.
    :returns: The list of loaded plugins initialized.
    """
    interfaces = []
    for interface in plugins.values():
        interfaces.append(
            interface['plugin'](interface['title'],
                                config))
    return interfaces


def import_resources(interfaces: list) -> dict:
    """
    Method to get all resources from the *input* interfaces.

    :param interfaces: The list of initialized input interfaces.
    :returns: The resources returned by the input interfaces.
    """
    resources = {}
    for interface in interfaces:
        resources.update(interface.import_resources())
    if not resources:
        raise NoResourcesFound("No resources found")
    return resources


def export_resources(interfaces: list, resources: dict) -> None:
    """
    Method to export the resources using the *output* interfaces.

    :param interfaces: The list of initialized output interfaces.
    :param resources: The resources provided by the input interfaces.
    """
    for interface in interfaces:
        interface.export_resources(resources)


def argument_parse() -> argparse.ArgumentParser:
    """
    Method to extract the arguments from the command line.

    :returns: The argument parser.
    """
    parser = argparse.ArgumentParser(
        description="Generates rundeck resources "
                    "file from different API sources.")

    parser.add_argument(
        'config', type=str,
        help='Configuration file.')
    parser.add_argument(
        '-v', '--verbose', action='count', default=0,
        help='Verbosity level to use.')
    parser.add_argument(
        '-l', '--logger', type=str,
        help='The logger YAML configuration file.')
    parser.add_argument(
        '--no-cache', action='store_true',
        help='Do not use cache.')
    parser.add_argument(
        '-V', '--version', action='version',
        version='%(prog)s {}'.format(__version__),
        help='Prints version.')
    return parser


def verbosity(verbose: int):
    """
    Method to set the verbosity.

    :param verbose: The verbosity set by user.
    :returns: The verbosity level.
    """
    if verbose == 0:
        return logging.ERROR
    elif verbose == 1:
        return logging.WARNING
    elif verbose == 2:
        return logging.INFO
    elif verbose > 2:
        return logging.DEBUG



import yaml
import logging
from .interfaces import ResourcesExporter
from .common import normalize_path, get_section


class YAMLExporter(ResourcesExporter):
    """
    The YAML rundeck exporter class
    """

    def __init__(self, title: str, config: dict = {}) -> None:
        """
        Initialize the *YAMLExporter plugin.

        :param title: The title provided by the configuration section.
        :param config: The configuration provided by `config.read_config`.
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.title = title
        self.section = get_section('YAMLExporter', title)
        self.config = config.get(self.section, None)

    def export_resources(self, dictionary: dict) -> None:
        """
        Method to save nodes' information into a rundeck
        nodes resources `YAML` file.

        :param dictionary: Rundeck formatted nodes information.
        """
        self.logger.info("Exporting Rundeck resources in YAML format")
        self.logger.debug("Getting resources export path")
        yaml_file = YAMLExporter.export_path(self.config)
        if yaml_file:
            self.logger.debug("Export path found, normalizing path")
            abs_path = normalize_path(yaml_file)
            self.logger.debug("Saving Rundeck YAML resources file")
            with open(abs_path, 'w+') as yfile:
                yaml.dump(dictionary,
                          stream=yfile,
                          explicit_start=True,
                          default_flow_style=False)
        else:
            self.logger.debug("Export path not found, dumping data to stdout")
            print(yaml.dump(dictionary,
                            explicit_start=True,
                            default_flow_style=False))

    @staticmethod
    def export_path(config: dict) -> str:
        """
        Method to get the rundeck export file in absolute
        path format.

        :param config: The `Rundeck` section of the configuration.
        :returns: The file path of rundeck resources to export to.
        """
        return config.get('export_path', '') if config else ''



import os
import glob
import json
import logging
import tempfile

from .errors import CacheNotFound


class Cache:
    """
    A cache module the plugins can take advantage of in case the APIs
    do not return data or the server is down.
    """

    def __init__(self, config: dict, no_cache: bool) -> None:
        """
        Initialize the *Cache* system for `rundeck_resources`.

        :param config: The configuration provided by `config.read_config`.
        :param no_cache: Configure whether to use the cache or not.
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.separator = '-'
        self.no_cache = no_cache
        self.cache_config = {} if no_cache else self.load_cache(config)

    def load_cache(self, config: dict) -> dict:
        """
        Method to search for all configured plugins cache files

        :param config: the configuration provided by `config.read_config`.
        :returns: The cache configuration that includes paths to all plugin
                  cache files.
        """
        self.logger.info("Loading cache configuration")
        tempdir = tempfile.gettempdir()
        self.logger.debug("The tempdir is '%s'", tempdir)
        rundeck_tempdir = glob.glob('{}/rundeck_resources*'
                                    .format(tempdir))
        cache = {}
        if rundeck_tempdir:
            self.logger.debug(
                "Cache directory found, searching for plugins cache files")
            cache = self._load_plugin_cache(rundeck_tempdir[0],
                                            config)
        else:
            self.logger.debug("No cache directory found")
            cache = self._create_cache_tempdir(config)
        return cache

    @staticmethod
    def translate_to_filename(plugin_name: str) -> str:
        """
        Static method to translate the plugin name to a standard prefix
        that could be used in a standard matter in the *Cache* class.

        :param plugin_name: The plugin to translate the name for.
        :returns: The translated plugin name.
        """
        return plugin_name.replace(':', '_').lower()

    def _create_cache_tempdir(self, config: dict) -> dict:
        """
        Private method to create the *rundeck-resources* cache directory.

        *NOTE*: This will initialize all configured plugins cache file paths.

        :param config: The dictionary provided by `config.read_config`.
        :returns: The cache configuration that includes the path to the
                  *rundeck-resources* cache directory.
        """
        self.logger.info("Creating cache directory")
        cache = {}
        cache['rundeck_resources_tempdir'] = \
            tempfile.mkdtemp(prefix='rundeck_resources{}'
                             .format(self.separator))
        self.logger.debug("Initializing an empty list of plugin cache paths")
        for plugin in config:
            cache[plugin] = ''
        return cache

    def _load_plugin_cache(self, rundeck_tempdir: str, config: dict) -> dict:
        """
        Private method that will search for all configured plugins cache files.

        :param rundeck_tempdir: The *rundeck-resources* cache directory.
        :param config: The configuration provided by `config.read_config`.
        :returns: The cache configuration with all plugins cache file paths.
        """
        self.logger.info("Searching for plugin cache files")
        cache = {}
        cache['rundeck_resources_tempdir'] = rundeck_tempdir
        rundeck_cache_files = glob.glob('{}/*'.format(rundeck_tempdir))
        for plugin in config.keys():
            cache_file_prefix = Cache.translate_to_filename(plugin)
            for plugin_cache_file in rundeck_cache_files:
                plugin_cache_filename = plugin_cache_file.split('/')[-1]
                if plugin_cache_filename.startswith(cache_file_prefix):
                    cache[plugin] = plugin_cache_file
                    self.logger.debug("Plugin '%s' cache file located at '%s'",
                                      plugin, plugin_cache_file)
                    break
            else:
                cache[plugin] = ''
                self.logger.debug("Plugin '%s' cache file not found",
                                  plugin)
        return cache

    def invalidate(self, plugin: str) -> None:
        """
        Method to invalidate a plugin cache file.

        :param plugin: The plugin name to invalidate the cache file for.
        """
        self.logger.info("Invalidating cache for plugin '%s'",
                         plugin)
        plugin_tempfile = self.cache_config.get(plugin, '')
        if plugin_tempfile:
            self.logger.debug("Plugin '%s' cache was found at '%s' attempting"
                              " to remove", plugin, plugin_tempfile)
            try:
                os.remove(plugin_tempfile)
            except FileNotFoundError as e:
                self.logger.error("Failed to remove '%s': %s",
                                  plugin_tempfile, e)
                pass
        self.cache_config[plugin] = ''

    def cache(self, plugin: str, resources: dict) -> None:
        """
        Method to cache plugin data.

        :param plugin: The plugin name to cache data for.
        :param resources: The resources data to cache.
        """
        if self.no_cache:
            self.logger.info("The cache is disabled, no caching to perform.")
        else:
            self.logger.info("Caching data for plugin '%s'", plugin)
            cache_tempdir = self.cache_config.get('rundeck_resources_tempdir')
            plugin_name_prefix = '{}{}'.format(
                Cache.translate_to_filename(plugin), self.separator)
            self.logger.debug("Creating new cache file for plugin '%s'",
                              plugin)
            fd, new_tempfile = tempfile.mkstemp(prefix=plugin_name_prefix,
                                                dir=cache_tempdir)
            self.logger.debug("Saving resources data for plugin '%s'"
                              " in new cache file", plugin)
            with open(new_tempfile, 'w') as f:
                f.write(json.dumps(resources))
            os.close(fd)
            self.invalidate(plugin)
            self.logger.debug("Updating '%s' plugin cache path with"
                              " the new cache file", plugin)
            self.cache_config[plugin] = new_tempfile

    def uncache(self, plugin: str) -> dict:
        """
        Method to read cached data for a specific plugin.

        :param plugin: The plugin to read cached data for.
        :raises: CacheNotFound
        """
        if self.no_cache:
            self.logger.info("The cache is disabled, returning an empty"
                             " dictionary")
            return {}
        self.logger.info("Retrieving cached data for plugin '%s'", plugin)
        plugin_tempfile = self.cache_config.get(plugin, '')
        if not plugin_tempfile:
            self.logger.warning("No cached data for plugin '%s' was found"
                                ", raising error", plugin)
            raise CacheNotFound("No cache found for plugin '{}'"
                                .format(plugin))

        self.logger.debug("Cached data for plugin '%s' found"
                          " and is being returned", plugin)
        with open(plugin_tempfile, 'r') as f:
            return json.loads(f.read())



import os
import yaml
import logging.config
from .common import check_file


def setup_logging(default_path: str = None,
                  default_level: int = logging.ERROR,
                  env_key: str = 'LOG_CFG') -> None:
    """
    Method that sets the logging system up.

    :param default_path: The path to the logger configuration.
    :param default_level: The default logging level (DEFAULT: ERROR)
    :param env_key: The environment variable specifying the path to the
                    configuration file.
    """
    value = os.getenv(env_key, None)
    path = None
    if default_path:
        try:
            path = check_file(default_path)
        except FileNotFoundError:
            path = None
    if value:
        try:
            path = check_file(value)
        except FileNotFoundError:
            path = None

    if path:
        with open(path, mode='r') as f:
            config = yaml.safe_load(f.read())
            logging.config.dictConfig(config)
    else:
        _format = '%(asctime)s - %(levelname)s - %(filename)s:' \
            '%(name)s.%(funcName)s:%(lineno)d - %(message)s'
        logging.basicConfig(level=default_level, format=_format)



import logging
import pkg_resources
from .errors import ConfigError

# Setup logging
logger = logging.getLogger(__name__)


def load_plugins(config: dict, resource_types: str) -> dict:
    """
    Method to load plugins defined in the *entry_point*
    by *resource_type*.

    :param config: The configuration file content.
    :param resource_types: The resource type to load (Input, Output).
    :returns: The plugins loaded.
    """
    logger.debug("Loading plugins configured to be used")
    raised_error = Exception()
    entry_points = get_plugins(resource_types)
    plugins = {}
    for section in config:
        _section = section.split(":")
        try:
            if _section[0] in entry_points:
                plugins.update({_section[0]: {
                    'plugin': entry_points[_section[0]].load(),
                    'title': _section[1]
                }})
        except IndexError as e:
            raised_error = e

        if isinstance(raised_error, IndexError):
            raise ConfigError("No plugin title found, in section '{}'."
                              " Configuration sections need to be in"
                              " the format 'PluginName:Title'"
                              .format(section))
    return plugins


def get_plugins(resource_types: str) -> dict:
    """
    Method to get the plugins defined in the *entry_point*.

    :param resource_types: The resource type to get (Input, Output).
    :returns: The entry points by name.
    """
    logger.debug("Importing plugins from list available")
    entry_points = {}
    for entry_point in pkg_resources.iter_entry_points(resource_types):
        entry_points.update({entry_point.name: entry_point})
    return entry_points



import logging
from chef import ChefAPI, Search

from .cache import Cache
from .interfaces import ResourcesImporter
from .common import check_file, get_section
from .errors import CacheNotFound


class ChefImporter(ResourcesImporter):
    """
    A Chef node information importer
    """

    def __init__(self, title: str, config: dict, cache: Cache) -> None:
        """
        Initialize the *ChefImporter* plugin.

        :param title: The title provided by the configuration section.
        :param config: The configuration provided by `config.read_config`.
        :param cache: The cache system instance.
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.cache = cache
        self.title = title
        self.section = get_section('ChefImporter', title)
        self.config = config[self.section]

    def get_chef_nodes(self) -> dict:
        """
        Method to get the chef nodes information.

        :returns: The chef nodes information.
        """
        self.logger.debug("Importing Chef resources")
        self.logger.debug("Expanding certificate paths")
        try:
            config = ChefImporter.expand_paths(self.config)
        except FileNotFoundError:
            self.logger.error(
                "Configuration file was not found, skipping this run "
                "and returning no data")
            return {}
        self.logger.debug("Querying Chef server API")
        chef_nodes = ChefImporter.call_chef(config)
        return chef_nodes

    @staticmethod
    def call_chef(config: dict) -> dict:
        """
        Method to query the chef server.

        :param config: The chef configuration.
        :returns: The chef nodes information.
        """
        with ChefAPI(config['url'],
                     config['user_cert_path'],
                     config['user'],
                     version=config.get('version', '0.10.8'),
                     ssl_verify=config.get('ssl_cert_path', False)):
            return Search('node', 'Node Name:*')

    @staticmethod
    def expand_paths(config: dict) -> dict:
        """
        Method to expand file paths for the user certificate
        and the chef ssl certificate.

        :param config: The chef configuration.
        :returns: The chef configuration with expanded paths.
        """
        config['user_cert_path'] = ChefImporter.expand_user_cert_path(
            config['user_cert_path'])
        config['ssl_cert_path'] = ChefImporter.expand_ssl_cert_path(
            config.get('ssl_cert_path', None))
        return config

    @staticmethod
    def expand_user_cert_path(user_cert_path: str) -> str:
        """
        Method to return expanded and checked `user_cert_path`.

        :param user_cert_path: The user certificate path.
        :returns: The check expanded user certificate path.
        """
        return check_file(user_cert_path)

    @staticmethod
    def expand_ssl_cert_path(ssl_cert_path: str) -> str:
        """
        Method to return expanded and checked `ssl_cert_path`.

        :param user_cert_path: The chef server SSL certificate path.
        :returns: The check expanded chef server SSL certificate path.
        """
        try:
            return check_file(ssl_cert_path)
        except FileNotFoundError:
            return ''

    def import_resources(self) -> dict:
        """
        Method to format chef resources into rundeck resources.

        :returns: Rundeck formatted nodes resources.
        """
        self.logger.info("Importing Rundeck resources from Chef plugin")
        chef_nodes = self.get_chef_nodes()
        nodes = {}
        self.logger.debug("Parsing Chef resources")
        try:
            for node in chef_nodes:
                try:
                    _node = {}
                    _node[node['name']] = {}
                    _node[node['name']]['hostname'] = node['automatic']['fqdn']
                    _node[node['name']]['nodename'] = \
                        node['automatic']['hostname']
                    _node[node['name']]['Environment'] = \
                        node['chef_environment']
                    _node[node['name']]['osName'] = \
                        node['automatic']['platform']
                    _node[node['name']]['osFamily'] = \
                        node['automatic']['platform_family']
                    _node[node['name']]['osVersion'] = \
                        node['automatic']['platform_version']
                    _node[node['name']]['osArch'] = \
                        node['automatic']['kernel']['machine']
                    _node[node['name']]['recipes'] = \
                        ','.join(node['automatic']['recipes'])
                    _node[node['name']]['roles'] = \
                        ','.join(node['automatic']['roles'])
                    _node[node['name']]['tags'] = \
                        ','.join(node['automatic']['recipes'])
                    _node[node['name']]['chef_tags'] = \
                        ','.join(node['normal']['tags'])
                    _description = \
                        node['automatic']['lsb'].get('description',
                                                     None)
                    if _description:
                        _node[node['name']]['description'] = \
                            _description
                    _username = self.config.get('rundeck_user_login', None)
                    if _username:
                        _node[node['name']]['username'] = _username
                except KeyError as e:
                    self.logger.warning("KeyError encountered with node '%s':"
                                        " %s",
                                        node['automatic']['hostname'], e)
                    continue
                nodes.update(_node)
            self.cache.cache(self.section, nodes)
        except AttributeError as e:
            self.logger.error("ChefAPI encountered an issues: %s", e)
            self.logger.warning("Retrieving data from cache")
            try:
                nodes.update(self.cache.uncache(self.section))
            except CacheNotFound as e:
                self.logger.warning(e)
                return nodes
        self.logger.debug("Returning Rundeck resources from Chef plugin")
        return nodes



import os
import logging

# Setup logging
logger = logging.getLogger(__name__)


def normalize_path(path: str) -> str:
    """
    Method to expand and return an absolute
    path from a normal path.

    :param path: The path to normalize.
    :returns: The absolute path.
    """
    logger.debug("Normalizing path: %s", path)
    exp_path = os.path.expanduser(path)
    abs_path = os.path.abspath(exp_path)
    logger.debug("Normalized path: %s", abs_path)
    return abs_path


def check_file(path: str) -> str:
    """
    Method to normalize the path of a file and
    check if the file exists and is a file.

    :param path: The file path to check.
    :returns: The absolute path of a file.
    :raises: FileNotFoundError
    """
    logger.debug("Checking file: %s", path)
    file_path = normalize_path(path)
    if not os.path.exists(file_path) or not os.path.isfile(file_path):
        logger.error("File '%s' not found, raising exception", file_path)
        raise FileNotFoundError
    logger.debug("File '%s' found, returning path", file_path)
    return file_path


def get_section(plugin: str, title: str) -> str:
    """
    Construct the configuration section

    :param plugin: The plugin name.
    :param title: The title seperated by `:` in the configuration file.
    :returns: The configuration section to pull from the configuration
                file.
    """
    return ':'.join([plugin, title])



import logging
import configparser
from .common import check_file

# Setup logging
logger = logging.getLogger(__name__)


def read_config(path: str) -> dict:
    """
    Method to read the init configuration file.

    :param path: The path to the init configuration file.
    :returns: The content of the configuration file.
    """
    logger.debug("Reading configuration file")
    config_path = check_file(path)
    logger.debug("Parsing configuration file content")
    config = configparser.ConfigParser()
    config.read(config_path)
    configuration = {}
    logger.debug("Converting configuration into a dictionary")
    for section in config.sections():
        configuration[section] = {}
        for key, val in config.items(section):
            configuration[section][key] = val
    logger.debug("Returning configuration dictionary")
    return configuration



from pathlib import Path
from setuptools import setup
from rundeck_resources import __version__

here = Path(__file__).absolute().parent

with open(str(here.joinpath('README.rst')), encoding='utf-8') as f:
    long_description = f.read()

with open(str(here.joinpath('requirements/requirements.txt')),
          encoding='utf-8') as f:
    install_requires = f.read()

description = 'Rundeck Resources Generator'

setup(
    name='rundeck_resources',
    version=__version__,
    license='BSD',
    author='Elia El Lazkani',
    author_email='eliaellazkani@gmail.com',
    url='https://gitlab.com/elazkani/rundeck-resources',
    python_requires='>=python3.5',
    description=description,
    long_description=long_description,
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'rundeck-resources = rundeck_resources.cli:main'
        ],
        'Importers': [
            'ChefImporter = rundeck_resources.chef_importer:ChefImporter'
        ],
        'Exporters': [
            'YAMLExporter = rundeck_resources.yaml_exporter:YAMLExporter'
        ]
    },
    packages=['rundeck_resources'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Environment :: Console',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators'
    ],
    zip_safe=False
)



class ConfigError(Exception):
    """
    Custom `ConfigError` exception.
    We raise this exception when a configuration error
    is encountered.
    """

    def __init__(self, message: str = None) -> None:
        super().__init__(message)


class NoResourcesFound(Exception):
    """
    Custom `NoResourcesFound` exception.
    We raise this exception when no resources have been
    returned from the importers.
    """

    def __init__(self, message: str = None) -> None:
        super().__init__(message)


class CacheNotFound(Exception):
    """
    Custom `CacheNotFound` exception.
    We raise this exception when no plugin cache have been
    found.
    """

    def __init__(self, message: str = None) -> None:
        super().__init__(message)
